from fastapi import FastAPI, Request, Depends
from fastapi.security import APIKeyHeader
from fastapi.exceptions import HTTPException
from fastapi.responses import JSONResponse
from pydantic import BaseModel
from typing import Any
import uvicorn
from hasura_ndc.connector import Connector, ConfigurationType, StateType
from hasura_ndc.models import (CapabilitiesResponse, SchemaResponse, QueryResponse, ExplainResponse, MutationResponse,
                               QueryRequest, MutationRequest, VERSION)
from opentelemetry import trace
import hasura_ndc.instrumentation as instrumentation
from hasura_ndc.errors import ConnectorError

tracer = trace.get_tracer("ndc-sdk-python.server")


class ServerOptions(BaseModel):
    configuration: str
    port: int
    service_token_secret: str
    log_level: str
    pretty_print_logs: str


async def start_server(connector: Connector[ConfigurationType, StateType],
                       options: ServerOptions):
    api_key_header = APIKeyHeader(name="Authorization", auto_error=False)
    configuration = await connector.parse_configuration(configuration_dir=options.configuration)
    metrics = {}
    state = await connector.try_init_state(configuration, metrics)

    async def get_api_key(header: str = Depends(api_key_header)):
        match_value = f"Bearer {options.service_token_secret}" if options.service_token_secret else None
        if header == match_value:
            return header
        else:
            raise HTTPException(status_code=403, detail="Invalid API Key")

    app = FastAPI()

    @app.get("/capabilities", dependencies=[Depends(get_api_key)])
    async def get_capabilities() -> CapabilitiesResponse:
        capabilities = await connector.get_capabilities(configuration)
        return CapabilitiesResponse(
            version=VERSION,
            capabilities=capabilities
        )

    @app.get("/health")
    async def get_health_readiness() -> Any:
        return await connector.get_health_readiness(configuration, state)

    @app.get("/metrics", dependencies=[Depends(get_api_key)])
    async def fetch_metrics() -> Any:
        return await connector.fetch_metrics(configuration, state)

    @app.get("/schema", dependencies=[Depends(get_api_key)])
    async def get_schema(request: Request) -> SchemaResponse:
        return await instrumentation.with_active_span(
            tracer,
            "getSchema",
            lambda span: connector.get_schema(configuration),
            headers=dict(request.headers)
        )

    @app.post("/query", dependencies=[Depends(get_api_key)])
    async def execute_query(request: Request) -> QueryResponse:
        headers = dict(request.headers)
        request_data = await request.json()
        return await instrumentation.with_active_span(
            tracer,
            "query",
            lambda span: connector.query(configuration, state, QueryRequest(**request_data)),
            headers=headers
        )

    @app.post("/query/explain", dependencies=[Depends(get_api_key)])
    async def query_explain(request: Request) -> ExplainResponse:
        headers = dict(request.headers)
        request_data = await request.json()
        return await instrumentation.with_active_span(
            tracer,
            "query_explain",
            lambda span: connector.query_explain(configuration, state, QueryRequest(**request_data)),
            headers=headers
        )

    @app.post("/mutation", dependencies=[Depends(get_api_key)])
    async def execute_mutation(request: Request) -> MutationResponse:
        headers = dict(request.headers)
        request_data = await request.json()
        return await instrumentation.with_active_span(
            tracer,
            "mutation",
            lambda span: connector.mutation(configuration, state, MutationRequest(**request_data)),
            headers=headers
        )

    @app.post("/mutation/explain", dependencies=[Depends(get_api_key)])
    async def mutation_explain(request: Request) -> ExplainResponse:
        headers = dict(request.headers)
        request_data = await request.json()
        return await instrumentation.with_active_span(
            tracer,
            "mutation_explain",
            lambda span: connector.mutation_explain(configuration, state, MutationRequest(**request_data)),
            headers=headers
        )

    @app.exception_handler(Exception)
    async def http_exception_handler(_: Request, e: Exception):
        if isinstance(e, ConnectorError):
            return JSONResponse(
                status_code=e.status_code,
                content={
                    "message": str(e.message),
                    "details": e.details if hasattr(e, 'details') else {}
                }
            )
        else:
            return JSONResponse(
                status_code=500,
                content={
                    "message": str(e),
                    "details": {}
                }
            )

    config = uvicorn.Config(app, host=None, port=options.port, loop="asyncio")
    server = uvicorn.Server(config)
    await server.serve()



from hasura_ndc.connector import Connector
from hasura_ndc.main import start
from hasura_ndc.models import *
from pydantic import BaseModel
import inspect
import asyncio
from typing import Optional, get_origin, get_args, Union
import types

class HeaderMap(dict):
    pass


class Configuration(BaseModel):
    pass


class State(BaseModel):
    pass


class FunctionConnector(Connector[Configuration, State]):

    def __init__(self):
        super().__init__(Configuration, State)
        self.query_functions = {}
        self.mutation_functions = {}

    async def parse_configuration(self, configuration_dir: str) -> Configuration:
        config = Configuration()
        return config

    async def try_init_state(self, configuration: Configuration, metrics: Any) -> State:
        return State()

    async def get_capabilities(self, configuration: Configuration) -> Capabilities:
        return Capabilities(
                query=QueryCapabilities(
                    aggregates=LeafCapability(),
                    variables=LeafCapability(),
                    explain=LeafCapability()
                ),
                mutation=MutationCapabilities(
                    transactional=LeafCapability(),
                    explain=None
                )
            )

    async def query_explain(self,
                            configuration: Configuration,
                            state: State,
                            request: QueryRequest) -> ExplainResponse:
        pass

    async def mutation_explain(self,
                               configuration: Configuration,
                               state: State,
                               request: MutationRequest) -> ExplainResponse:
        pass

    async def fetch_metrics(self,
                            configuration: Configuration,
                            state: State) -> Optional[None]:
        pass

    async def get_health_readiness(self,
                           configuration: Configuration,
                           state: State) -> Optional[None]:
        pass

    async def get_schema(self, configuration: Configuration) -> SchemaResponse:
        functions = []
        procedures = []
        scalar_types = {
                "String": ScalarType(
                    representation=StringType(type="string"),
                    aggregate_functions={},
                    comparison_operators={}
                ),
                "Int": ScalarType(
                    representation=IntegerType(type="integer"),
                    aggregate_functions={},
                    comparison_operators={}
                ),
                "Float": ScalarType(
                    representation=Float64Type(type="float64"),
                    aggregate_functions={},
                    comparison_operators={}
                ),
                "Boolean": ScalarType(
                    representation=BooleanType(type="boolean"),
                    aggregate_functions={},
                    comparison_operators={}
                ),
                "Json": ScalarType(
                    representation=JsonType(type="json"),
                    aggregate_functions={},
                    comparison_operators={}
                ),
                "HeaderMap": ScalarType(
                    representation=JsonType(type="json"),
                    aggregate_functions={},
                    comparison_operators={}
                )
            }
        object_types = {}
        for name, (func, _parallel_degree) in self.query_functions.items():
            function_info = self.generate_function_info(name, func, object_types)
            functions.append(function_info)

        for name, func in self.mutation_functions.items():
            procedure_info = self.generate_procedure_info(name, func, object_types)
            procedures.append(procedure_info)

        schema_response = SchemaResponse(
            scalar_types=scalar_types,
            functions=functions,
            procedures=procedures,
            object_types=object_types,
            collections=[]
        )

        return schema_response

    def generate_function_info(self, name, func, object_types):
        signature = inspect.signature(func)
        arguments = {}
        for arg_name, arg_type in signature.parameters.items():
            if arg_type.annotation != HeaderMap:
                arguments[arg_name] = {
                    "type": self.get_type_info(arg_type, name, object_types, arg_name)
                }
        
        description = None
        if func.__doc__:
            description = func.__doc__.strip()

        return FunctionInfo(
            name=name,
            arguments=arguments,
            result_type=self.get_type_info(signature.return_annotation, name, object_types, None),
            description=description
        )

    def generate_procedure_info(self, name, func, object_types):
        signature = inspect.signature(func)
        arguments = {}
        for arg_name, arg_type in signature.parameters.items():
            if arg_type.annotation != HeaderMap:
                arguments[arg_name] = {
                    "type": self.get_type_info(arg_type, name, object_types, arg_name)
                }

        description = None
        if func.__doc__:
            description = func.__doc__.strip()
        
        return ProcedureInfo(
            name=name,
            arguments=arguments,
            result_type=self.get_type_info(signature.return_annotation, name, object_types, None),
            description=description
        )

    @staticmethod
    def get_type_info(typ, caller_name, object_types, arg_name=None):
        if isinstance(typ, inspect.Parameter):
            typ = typ.annotation
        # If the type is an array we should return a ArrayType with the underlying type
        if typ == int:
            res = NamedType(type="named", name="Int")
        elif typ == float:
            res = NamedType(type="named", name="Float")
        elif typ == str:
            res = NamedType(type="named", name="String")
        elif typ == bool:
            res = NamedType(type="named", name="Boolean")
        elif typ == inspect._empty:
            res = NamedType(type="named", name="Json")
        elif typ == list or get_origin(typ) == list:
            if typ == list:
                res = ArrayType(type="array", element_type=NamedType(type="named", name="Json"))              
            elif len(typ.__args__) == 0:
                res = ArrayType(type="array", element_type=NamedType(type="named", name="Json"))
            else:
                model_name = f"{caller_name}"
                if arg_name:
                    model_name = f"{caller_name}_{arg_name}"
                res = ArrayType(type="array", element_type=FunctionConnector.get_type_info(typ.__args__[0], model_name, object_types, None))
        elif get_origin(typ) in (Union, types.UnionType) and type(None) in get_args(typ):
            args = get_args(typ)
            non_none_types = [t for t in args if t != type(None)]
            if len(non_none_types) == 1:
                model_name = f"{caller_name}"
                if arg_name:
                    model_name = f"{caller_name}_{arg_name}"
                res = NullableType(type="nullable", underlying_type=FunctionConnector.get_type_info(non_none_types[0], model_name, object_types, None))
            else:
                res = NullableType(type="nullable", underlying_type=NamedType(type="named", name="Json"))
        elif issubclass(typ, BaseModel):
            model_name = f"{caller_name}"
            if arg_name:
                model_name = f"{caller_name}_{arg_name}"
            if model_name not in object_types:
                fields = {}
                for name, field in typ.model_fields.items():
                    field_type = FunctionConnector.get_type_info(field.annotation, model_name, object_types, name)
                    fields[name] = ObjectField(
                        type=field_type,
                        description=field.description
                    )
                object_types[model_name] = ObjectType(
                    description=typ.__doc__,
                    fields=fields
                )
            res = NamedType(type="named", name=model_name)
        else:
            res = NamedType(type="named", name="Json")
        return res

    @staticmethod
    def reshape_result(result, requested_fields):
        def reshape_value(value, fields):
            if value is None:
                return None

            if isinstance(fields, NestedArray):
                if isinstance(value, list):
                    return [reshape_value(item, fields.fields) for item in value]
                else:
                    return reshape_value(value, fields.fields)
            elif isinstance(fields, NestedObject):
                return reshape_object(value, fields.fields)
            elif isinstance(fields, ColumnField):
                if fields.fields:
                    return reshape_value(getattr(value, fields.column), fields.fields)
                else:
                    return getattr(value, fields.column)
            else:
                return value

        def reshape_object(obj, fields):
            if obj is None:
                return None

            reshaped = {}
            for field_alias, field in fields.items():
                if isinstance(field, ColumnField):
                    value = getattr(obj, field.column)
                    if field.fields:
                        reshaped[field_alias] = reshape_value(value, field.fields)
                    else:
                        reshaped[field_alias] = value
                elif isinstance(field, (NestedObject, NestedArray)):
                    reshaped[field_alias] = reshape_value(getattr(obj, field_alias), field)
            return reshaped

        return reshape_value(result, requested_fields)

    @staticmethod
    def cast_to_type(value, target_type):
        origin = get_origin(target_type)
        if origin == types.UnionType:
            args = get_args(target_type)
            if types.NoneType in args and len(args) == 2:
                # target type is just a nullable type
                if value is None:
                    return None
                other_type = next(t for t in args if t is not types.NoneType)
                return FunctionConnector.cast_to_type(value, other_type)
        elif origin == list:
            args = get_args(target_type)
            if len(args) == 1:
                element_type = args[0]
                return [FunctionConnector.cast_to_type(item, element_type) for item in value]
            else:
                return value  # If it's List without type argument, return as is
        elif origin is not None and issubclass(origin, BaseModel):
            return origin(**value)
        elif inspect.isclass(target_type) and issubclass(target_type, BaseModel):
            return target_type(**value)
        else:
            return value


    async def query(self, configuration: Configuration, state: State, request: QueryRequest) -> QueryResponse:
        func, parallel_degree = self.query_functions[request.collection]
        signature = inspect.signature(func)
        
        root_args = {}
        root_vars = {}
        for k, v in request.arguments.items():
            if v.type == "literal":
                arg_type = signature.parameters[k].annotation
                if isinstance(arg_type, type) or (get_origin(arg_type) is not None):
                    root_args[k] = self.cast_to_type(v.value, arg_type)
                else:
                    root_args[k] = v.value
            elif v.type == "variable":
                root_vars[k] = v.name

        args_array = []
        if request.variables:
            for var in request.variables:
                var_args = {}
                for var_key, var_name in root_vars.items():
                    arg_type = signature.parameters[var_key].annotation
                    var_args[var_key] = self.cast_to_type(var[var_name], arg_type)
                args_array.append({
                    **var_args,
                    **root_args
                })
        else:
            args_array = [root_args]


        async def process_args(args):
            if asyncio.iscoroutinefunction(func):
                result = await func(**args)
            else:
                result = func(**args)

            reshaped_result = FunctionConnector.reshape_result(result, request.query.fields["__value"].fields)
            
            return RowSet(
                aggregates=None,
                rows=[
                    {"__value": reshaped_result}
                ]
            )

        async def process_batch(batch):
            return await asyncio.gather(*[process_args(args) for args in batch])

        row_sets = []
        for i in range(0, len(args_array), parallel_degree):
            batch = args_array[i:i + parallel_degree]
            batch_results = await process_batch(batch)
            row_sets.extend(batch_results)

        return row_sets

    async def mutation(self, 
                       configuration: Configuration,
                       state: State,
                       request: MutationRequest) -> MutationResponse:
        responses = []
        for operation in request.operations:
            operation_name = operation.name
            func = self.mutation_functions[operation_name]
            signature = inspect.signature(func)
            args = {}
            if operation.arguments:
                for k, v in operation.arguments.items():
                    arg_type = signature.parameters[k].annotation
                    args[k] = self.cast_to_type(v, arg_type)
            if asyncio.iscoroutinefunction(func):
                response = await func(**args)
            else:
                response = func(**args)
            reshaped_response = FunctionConnector.reshape_result(response, operation.fields)
            responses.append(reshaped_response)
        return MutationResponse(
            operation_results=[
                MutationOperationResults(
                    type="procedure",
                    result=response
                ) for response in responses
            ]
        )

    def register_query(self, func=None, *, parallel_degree=1):
        
        def decorator(f):
            self.query_functions[f.__name__] = (f, parallel_degree)
            return f
        
        if func is None:
            return decorator
        else:
            return decorator(func)

    def register_mutation(self, func):
        self.mutation_functions[func.__name__] = func
        return func


from typing import Optional, Dict, Any

class ConnectorError(Exception):
    def __init__(self, status_code: int, message: str, details: Optional[Dict[str, Any]] = None):
        super().__init__(message)
        self.status_code = status_code
        self.message = message
        self.details = details or {}

class BadRequest(ConnectorError):
    """The request did not match the data connector's expectation based on this specification"""
    def __init__(self, message: str, details: Optional[Dict[str, Any]] = None):
        super().__init__(400, message, details)

class Forbidden(ConnectorError):
    """The request could not be handled because a permission check failed - for example, a mutation might fail because a check constraint was not met"""
    def __init__(self, message: str, details: Optional[Dict[str, Any]] = None):
        super().__init__(403, message, details)

class Conflict(ConnectorError):
    """The request could not be handled because it would create a conflicting state for the data source - for example, a mutation might fail because a foreign key constraint was not met"""
    def __init__(self, message: str, details: Optional[Dict[str, Any]] = None):
        super().__init__(409, message, details)

class UnprocessableContent(ConnectorError):
    """The request could not be handled because, while the request was well-formed, it was not semantically correct. For example, a value for a custom scalar type was provided, but with an incorrect type"""
    def __init__(self, message: str, details: Optional[Dict[str, Any]] = None):
        super().__init__(422, message, details)

class InternalServerError(ConnectorError):
    """The request could not be handled because of an error on the server"""
    def __init__(self, message: str, details: Optional[Dict[str, Any]] = None):
        super().__init__(500, message, details)

class NotSupported(ConnectorError):
    """The request could not be handled because it relies on an unsupported capability. Note: this ought to indicate an error on the caller side, since the caller should not generate requests which are incompatible with the indicated capabilities"""
    def __init__(self, message: str, details: Optional[Dict[str, Any]] = None):
        super().__init__(501, message, details)

class BadGateway(ConnectorError):
    """The request could not be handled because an upstream service was unavailable or returned an unexpected response, e.g., a connection to a database server failed"""
    def __init__(self, message: str, details: Optional[Dict[str, Any]] = None):
        super().__init__(502, message, details)


import os
import argparse
import asyncio
from hasura_ndc.connector import *
from hasura_ndc.server import start_server
import hasura_ndc.instrumentation as instrumentation

if not instrumentation.is_initialized():
    instrumentation.init_telemetry()


async def start_server_wrapper(connector, args):
    args.configuration = os.getenv('HASURA_CONFIGURATION_DIRECTORY', args.configuration)
    args.port = int(os.getenv('HASURA_CONNECTOR_PORT', args.port))
    args.service_token_secret = os.getenv('HASURA_SERVICE_TOKEN_SECRET', args.service_token_secret)
    args.log_level = os.getenv('HASURA_LOG_LEVEL', args.log_level)
    args.pretty_print_logs = bool(os.getenv('HASURA_PRETTY_PRINT_LOGS', args.pretty_print_logs))
    await start_server(connector, args)


def start(connector: Connector):
    parser = argparse.ArgumentParser(description='Connector CLI')
    subparsers = parser.add_subparsers(dest='command', required=True)

    serve_parser = subparsers.add_parser('serve', help='Start the server')
    _add_serve_command_arguments(serve_parser)
    serve_parser.set_defaults(func=start_server_wrapper)

    args = parser.parse_args()
    asyncio.run(args.func(connector, args))


def _add_serve_command_arguments(parser):
    # Setting defaults to None to allow for checks later
    parser.add_argument('--configuration', default=None, help='Path to the configuration file')
    parser.add_argument('--port', type=int, default=8080, help='Port number')
    parser.add_argument('--service-token-secret', default=None, help='Service token secret')
    parser.add_argument('--log-level', default='info', help='Logging level')
    parser.add_argument('--pretty-print-logs', default=False, help="Pretty print logs")



from typing import Generic, TypeVar, Any
from typing import Optional
from abc import ABC, abstractmethod
from hasura_ndc.models import (
    Capabilities,
    SchemaResponse,
    QueryRequest,
    QueryResponse,
    ExplainResponse,
    MutationRequest,
    MutationResponse
)
from pydantic import BaseModel

ConfigurationType = TypeVar('ConfigurationType', bound=BaseModel)
StateType = TypeVar('StateType', bound=BaseModel)


class Connector(ABC, Generic[ConfigurationType, StateType]):

    def __init__(self, configuration_type, state_type):
        self.configuration_type = configuration_type
        self.state_type = state_type

    @abstractmethod
    async def parse_configuration(self, configuration_dir: str) -> ConfigurationType:
        pass

    @abstractmethod
    async def try_init_state(self, configuration: ConfigurationType, metrics: Any) -> StateType:
        pass

    @abstractmethod
    async def fetch_metrics(self, configuration: ConfigurationType, state: StateType) -> Optional[Any]:
        pass

    @abstractmethod
    async def get_health_readiness(self, configuration: ConfigurationType, state: StateType) -> Optional[Any]:
        pass

    @abstractmethod
    def get_capabilities(self, configuration: ConfigurationType) -> Capabilities:
        pass

    @abstractmethod
    async def get_schema(self, configuration: ConfigurationType) -> SchemaResponse:
        pass

    @abstractmethod
    async def query_explain(self,
                            configuration: ConfigurationType,
                            state: StateType,
                            request: QueryRequest) -> ExplainResponse:
        pass

    @abstractmethod
    async def mutation_explain(self,
                               configuration: ConfigurationType,
                               state: StateType,
                               request: MutationRequest) -> ExplainResponse:
        pass

    @abstractmethod
    async def mutation(self, configuration: ConfigurationType, state: StateType,
                       request: MutationRequest) -> MutationResponse:
        pass

    @abstractmethod
    async def query(self, configuration: ConfigurationType, state: StateType, request: QueryRequest) -> QueryResponse:
        pass



from opentelemetry import trace, metrics
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter as GRPCTraceExporter
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter as GRPCMetricExporter
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter as HTTPTraceExporter
from opentelemetry.exporter.otlp.proto.http.metric_exporter import OTLPMetricExporter as HTTPMetricExporter
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from opentelemetry.instrumentation.logging import LoggingInstrumentor
from opentelemetry.instrumentation.requests import RequestsInstrumentor
from opentelemetry.propagate import set_global_textmap
from opentelemetry.propagators.composite import CompositePropagator
from opentelemetry.propagators.b3 import B3Format
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator
from opentelemetry.baggage.propagation import W3CBaggagePropagator
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace import Status, StatusCode
from opentelemetry.propagate import get_global_textmap
from enum import Enum
import asyncio
import os

sdk = None

USER_VISIBLE_SPAN_ATTRIBUTE = {
    "internal.visibility": "user",
}


class Protocol(str, Enum):
    GRPC = "grpc"
    HTTP_PROTOBUF = "http/protobuf"


def get_exporters(protocol: Protocol, endpoint: str):
    """Get the appropriate exporters based on the protocol."""
    if protocol == Protocol.GRPC:
        trace_exporter = GRPCTraceExporter(endpoint=endpoint)
        metric_exporter = GRPCMetricExporter(endpoint=endpoint)
    elif protocol == Protocol.HTTP_PROTOBUF:
        # For HTTP/protobuf, we need to append the specific endpoints
        trace_endpoint = f"{endpoint}/v1/traces"
        metric_endpoint = f"{endpoint}/v1/metrics"
        trace_exporter = HTTPTraceExporter(endpoint=trace_endpoint)
        metric_exporter = HTTPMetricExporter(endpoint=metric_endpoint)
    else:
        raise ValueError(f"Unsupported protocol: {protocol}")

    return trace_exporter, metric_exporter


def init_telemetry(default_service_name="hasura-ndc",
                   default_endpoint="http://localhost:4317",
                   default_protocol="grpc"):
    global sdk

    if is_initialized():
        raise Exception("Telemetry has already been initialized!")

    service_name = os.environ.get("OTEL_SERVICE_NAME", default_service_name)
    endpoint = os.environ.get("OTEL_EXPORTER_OTLP_ENDPOINT", default_endpoint)
    protocol = Protocol(os.environ.get("OTEL_EXPORTER_OTLP_PROTOCOL", default_protocol))

    resource = Resource.create(attributes={"service.name": service_name})

    trace_exporter, metric_exporter = get_exporters(protocol, endpoint)

    tracer_provider = TracerProvider(resource=resource)
    trace.set_tracer_provider(tracer_provider)

    span_processor = BatchSpanProcessor(trace_exporter)
    tracer_provider.add_span_processor(span_processor)

    metric_reader = PeriodicExportingMetricReader(metric_exporter)
    metric_provider = MeterProvider(metric_readers=[metric_reader])
    metrics.set_meter_provider(metric_provider)

    FastAPIInstrumentor().instrument(
        server_request_hook=lambda span, request: span.set_attributes(USER_VISIBLE_SPAN_ATTRIBUTE)
    )
    RequestsInstrumentor().instrument(
        tracer_provider=trace.get_tracer_provider(),
        span_callback=lambda span, request, response: span.set_attributes(USER_VISIBLE_SPAN_ATTRIBUTE),
    )
    LoggingInstrumentor().instrument(
        log_hook=lambda span, record, _: record.update(
            {
                "resource.service.name": service_name,
                "parent_span_id": span.get_span_context().span_id,
            }
        )
    )

    propagators = CompositePropagator(
        [
            TraceContextTextMapPropagator(),
            W3CBaggagePropagator(),
            B3Format(),
        ]
    )
    set_global_textmap(propagators)

    sdk = trace.get_tracer_provider()


def is_initialized():
    return sdk is not None


async def with_active_span(tracer, name, func, attributes=None, headers: dict = None):
    if headers is None:
        headers = {}
    if attributes is None:
        attributes = {}

    attributes = {**USER_VISIBLE_SPAN_ATTRIBUTE, **attributes}

    # Extract the trace context from the headers
    ctx = get_global_textmap().extract(headers)

    # Get the current span context
    current_span = trace.get_current_span()
    if current_span.is_recording():
        ctx = trace.set_span_in_context(current_span, ctx)

    def handle_error(span, err):
        if isinstance(err, Exception) or isinstance(err, str):
            span.record_exception(err)
        span.set_status(Status(StatusCode.ERROR))

    # Start a new span, passing the extracted context
    with tracer.start_as_current_span(name, attributes=attributes, context=ctx) as span:
        try:
            retval = func(span)
            if asyncio.iscoroutine(retval):
                retval = await retval
            return retval
        except Exception as e:
            handle_error(span, e)
            raise
        finally:
            span.end()



from setuptools import setup, find_packages

# Read the contents of your README file
with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='hasura_ndc',
    version='0.36',
    packages=find_packages(),
    install_requires=[
        # This line reads the requirements from your `requirements.txt`
        line.strip() for line in open('requirements.txt', 'r').readlines()
    ],
    author='Tristen Harr',
    author_email='tristen.harr@hasura.io',
    description='A Hasura Data Connector SDK',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    license='Apache License 2.0'
)



from __future__ import annotations
from typing import List, Dict, Optional, Literal, Any
from pydantic import BaseModel
from pydantic import Field as PydanticField
from typing_extensions import Annotated
import warnings

warnings.filterwarnings("ignore", message="Field name .* shadows an attribute in parent .*")


class RootModel(BaseModel):

    def dict(self, **kwargs):
        return super().model_dump(**kwargs, exclude_none=True)


class BooleanType(RootModel):
    type: Literal["boolean"]


class StringType(RootModel):
    type: Literal["string"]


class NumberType(RootModel):
    type: Literal["number"]


class IntegerType(RootModel):
    type: Literal["integer"]


class Int8Type(RootModel):
    type: Literal["int8"]


class Int16Type(RootModel):
    type: Literal["int16"]


class Int32Type(RootModel):
    type: Literal["int32"]


class Int64Type(RootModel):
    type: Literal["int64"]


class Float32Type(RootModel):
    type: Literal["float32"]


class Float64Type(RootModel):
    type: Literal["float64"]


class BigIntegerType(RootModel):
    type: Literal["biginteger"]


class BigDecimalType(RootModel):
    type: Literal["bigdecimal"]


class UUIDType(RootModel):
    type: Literal["uuid"]


class DateType(RootModel):
    type: Literal["date"]


class TimestampType(RootModel):
    type: Literal["timestamp"]


class TimestamptzType(RootModel):
    type: Literal["timestamptz"]


class GeographyType(RootModel):
    type: Literal["geography"]


class GeometryType(RootModel):
    type: Literal["geometry"]


class BytesType(RootModel):
    type: Literal["bytes"]


class JsonType(RootModel):
    type: Literal["json"]


class EnumType(RootModel):
    type: Literal["enum"]
    one_of: List[str]


TypeRepresentation = Annotated[
    BooleanType | StringType | NumberType | IntegerType | Int8Type | Int16Type | Int32Type | Int64Type |
    Float32Type | Float64Type | BigIntegerType | BigDecimalType | UUIDType | DateType | TimestampType |
    TimestamptzType | GeographyType | GeometryType | BytesType | JsonType | EnumType,
    PydanticField(discriminator='type')
]


class NamedType(RootModel):
    type: Literal["named"]
    name: str


class NullableType(RootModel):
    type: Literal["nullable"]
    underlying_type: 'Type'


class ArrayType(RootModel):
    type: Literal["array"]
    element_type: 'Type'


class PredicateType(RootModel):
    type: Literal["predicate"]
    object_type_name: str


Type = Annotated[
    NamedType | NullableType | ArrayType | PredicateType,
    PydanticField(discriminator='type')
]


class ColumnCountType(RootModel):
    type: Literal["column_count"]
    column: str
    field_path: Optional[List[str]] = None
    distinct: bool


class SingleColumnType(RootModel):
    type: Literal["single_column"]
    column: str
    field_path: Optional[List[str]] = None
    function: str


class StarCountType(RootModel):
    type: Literal["star_count"]


Aggregate = Annotated[
    ColumnCountType | SingleColumnType | StarCountType,
    PydanticField(discriminator='type')
]


class VariableArgument(RootModel):
    type: Literal["variable"]
    name: str


class LiteralArgument(RootModel):
    type: Literal["literal"]
    value: Any


Argument = Annotated[
    VariableArgument | LiteralArgument,
    PydanticField(discriminator='type')
]


class ColumnTarget(RootModel):
    type: Literal["column"]
    name: str
    field_path: Optional[List[str]] = None
    path: List['PathElement']


class RootCollectionColumnTarget(RootModel):
    type: Literal["root_collection_column"]
    name: str
    field_path: Optional[List[str]] = None


ComparisonTarget = Annotated[
    ColumnTarget | RootCollectionColumnTarget,
    PydanticField(discriminator='type')
]


class ColumnValue(RootModel):
    type: Literal["column"]
    column: 'ComparisonTarget'  # Forward reference to ComparisonTarget


class ScalarValue(RootModel):
    type: Literal["scalar"]
    value: Any


class VariableValue(RootModel):
    type: Literal["variable"]
    name: str


ComparisonValue = Annotated[
    ColumnValue | ScalarValue | VariableValue,
    PydanticField(discriminator='type')
]

# UnaryComparisonOperator = "is_null"
UnaryComparisonOperator = Literal["is_null"]


class AndExpression(RootModel):
    type: Literal["and"]
    expressions: List['Expression']


class OrExpression(RootModel):
    type: Literal["or"]
    expressions: List['Expression']


class NotExpression(RootModel):
    type: Literal["not"]
    expression: 'Expression'


class UnaryComparisonOperatorExpression(RootModel):
    type: Literal["unary_comparison_operator"]
    column: 'ComparisonTarget'
    operator: 'UnaryComparisonOperator'


class BinaryComparisonOperatorExpression(RootModel):
    type: Literal["binary_comparison_operator"]
    column: 'ComparisonTarget'
    operator: str
    value: 'ComparisonValue'


class ExistsExpression(RootModel):
    type: Literal["exists"]
    in_collection: 'ExistsInCollection'
    predicate: Optional['Expression'] = None


Expression = Annotated[
    AndExpression | OrExpression | NotExpression | UnaryComparisonOperatorExpression | BinaryComparisonOperatorExpression | ExistsExpression,
    PydanticField(discriminator='type')
]


class PathElement(RootModel):
    relationship: str
    arguments: Dict[str, 'RelationshipArgument']
    predicate: Optional['Expression'] = None


OrderDirection = Literal["asc", "desc"]


class ColumnOrderByTarget(RootModel):
    type: Literal["column"]
    name: str
    field_path: Optional[List[str]] = None
    path: List[PathElement]


class SingleColumnAggregateOrderByTarget(RootModel):
    type: Literal["single_column_aggregate"]
    column: str
    field_path: Optional[List[str]] = None
    function: str
    path: List[PathElement]


class StarCountAggregateOrderByTarget(RootModel):
    type: Literal["star_count_aggregate"]
    path: List[PathElement]


OrderByTarget = Annotated[
    ColumnOrderByTarget | SingleColumnAggregateOrderByTarget | StarCountAggregateOrderByTarget,
    PydanticField(discriminator='type')
]


class OrderBy(RootModel):
    elements: List['OrderByElement']


class OrderByElement(RootModel):
    order_direction: 'OrderDirection'
    target: 'OrderByTarget'


class Query(RootModel):
    aggregates: Optional[Dict[str, 'Aggregate']] = None
    fields: Optional[Dict[str, 'Field']] = None
    limit: Optional[int] = None
    offset: Optional[int] = None
    order_by: Optional['OrderBy'] = None
    predicate: Optional['Expression'] = None


class VariableRelationshipArgument(RootModel):
    type: Literal["variable"]
    name: str


class LiteralRelationshipArgument(RootModel):
    type: Literal["literal"]
    value: Any


class ColumnRelationshipArgument(RootModel):
    type: Literal["column"]
    name: str


RelationshipArgument = Annotated[
    VariableRelationshipArgument | LiteralRelationshipArgument | ColumnRelationshipArgument,
    PydanticField(discriminator='type')
]


class NestedObject(RootModel):
    type: Literal["object"]
    fields: Dict[str, Field]


class NestedArray(RootModel):
    type: Literal["array"]
    fields: 'NestedField'  # Forward reference to NestedField


NestedField = Annotated[
    NestedObject | NestedArray,
    PydanticField(discriminator='type')
]


class ColumnField(RootModel):
    type: Literal["column"]
    column: str
    fields: Optional['NestedField'] = None
    arguments: Optional[Dict[str, 'Argument']] = None


class RelationshipField(RootModel):
    type: Literal["relationship"]
    query: 'Query'
    relationship: str
    arguments: Dict[str, 'RelationshipArgument']


Field = Annotated[
    ColumnField | RelationshipField,
    PydanticField(discriminator='type')
]


class RelatedExistsInCollection(RootModel):
    type: Literal["related"]
    relationship: str
    arguments: Dict[str, 'RelationshipArgument']


class UnrelatedExistsInCollection(RootModel):
    type: Literal["unrelated"]
    collection: str
    arguments: Dict[str, 'RelationshipArgument']


class NestedCollectionExistsInCollection(RootModel):
    type: Literal["nested_collection"]
    column_name: str
    arguments: Optional[Dict[str, Argument]]
    field_path: Optional[List[str]]


ExistsInCollection = Annotated[
    RelatedExistsInCollection | UnrelatedExistsInCollection | NestedCollectionExistsInCollection,
    PydanticField(discriminator='type')
]

RelationshipType = Literal["object", "array"]

RowFieldValue = Any


class RowSet(RootModel):
    aggregates: Optional[Dict[str, Any]] = None
    rows: Optional[List[Dict[str, RowFieldValue]]] = None


QueryResponse = List[RowSet]


class Relationship(RootModel):
    column_mapping: Dict[str, str]
    relationship_type: 'RelationshipType'
    target_collection: str
    arguments: Dict[str, 'RelationshipArgument']


class MutationOperation(RootModel):
    type: Literal["procedure"]
    name: str
    arguments: Dict[str, Any]
    fields: Optional['NestedField'] = None


class MutationOperationResults(RootModel):
    type: Literal["procedure"]
    result: Any


class CapabilitiesResponse(RootModel):
    version: str
    capabilities: 'Capabilities'


class Capabilities(RootModel):
    query: 'QueryCapabilities'
    mutation: 'MutationCapabilities'
    relationships: Optional['RelationshipCapabilities'] = None


class SchemaRoot(RootModel):
    capabilities_response: CapabilitiesResponse
    schema_response: 'SchemaResponse'
    query_request: 'QueryRequest'
    query_response: 'QueryResponse'
    mutation_request: 'MutationRequest'
    mutation_response: 'MutationResponse'
    explain_response: 'ExplainResponse'
    error_response: 'ErrorResponse'
    validate_response: 'ValidateResponse'


class LeafCapability(RootModel):
    pass


class NestedFieldCapabilities(RootModel):
    filter_by: Optional['LeafCapability'] = None
    order_by: Optional['LeafCapability'] = None
    aggregates: Optional['LeafCapability'] = None


class ExistsCapabilities(RootModel):
    nested_collections: Optional['LeafCapability'] = None


class QueryCapabilities(RootModel):
    aggregates: Optional['LeafCapability'] = None
    variables: Optional['LeafCapability'] = None
    explain: Optional['LeafCapability'] = None
    nested_fields: Optional['NestedFieldCapabilities'] = None
    exists: Optional['ExistsCapabilities'] = None


class MutationCapabilities(RootModel):
    transactional: Optional['LeafCapability'] = None
    explain: Optional['LeafCapability'] = None


class RelationshipCapabilities(RootModel):
    relation_comparisons: Optional['LeafCapability'] = None
    order_by_aggregate: Optional['LeafCapability'] = None


class SchemaResponse(RootModel):
    scalar_types: Dict[str, 'ScalarType']
    object_types: Dict[str, 'ObjectType']
    collections: List['CollectionInfo']
    functions: List['FunctionInfo']
    procedures: List['ProcedureInfo']


class EqualOperator(RootModel):
    type: Literal["equal"]


class InOperator(RootModel):
    type: Literal["in"]


class CustomOperator(RootModel):
    type: Literal["custom"]
    argument_type: Optional['Type']


ComparisonOperatorDefinition = Annotated[
    EqualOperator | InOperator | CustomOperator,
    PydanticField(discriminator='type')
]


class ScalarType(RootModel):
    representation: Optional['TypeRepresentation'] = None
    aggregate_functions: Dict[str, 'AggregateFunctionDefinition']
    comparison_operators: Dict[str, 'ComparisonOperatorDefinition']


class AggregateFunctionDefinition(RootModel):
    result_type: 'Type'


class ObjectType(RootModel):
    description: Optional[str] = None
    fields: Dict[str, 'ObjectField']


class ObjectField(RootModel):
    description: Optional[str] = None
    type: 'Type'
    arguments: Optional[Dict[str, 'ArgumentInfo']] = None


class CollectionInfo(RootModel):
    name: str
    description: Optional[str] = None
    arguments: Dict[str, 'ArgumentInfo']
    type: str
    uniqueness_constraints: Dict[str, 'UniquenessConstraint']
    foreign_keys: Dict[str, 'ForeignKeyConstraint']


class ArgumentInfo(RootModel):
    description: Optional[str] = None
    type: 'Type'


class UniquenessConstraint(RootModel):
    unique_columns: List[str]


class ForeignKeyConstraint(RootModel):
    column_mapping: Dict[str, str]
    foreign_collection: str


class FunctionInfo(RootModel):
    name: str
    description: Optional[str] = None
    arguments: Dict[str, 'ArgumentInfo']
    result_type: 'Type'


class ProcedureInfo(RootModel):
    name: str
    description: Optional[str] = None
    arguments: Dict[str, 'ArgumentInfo']
    result_type: 'Type'


class QueryRequest(RootModel):
    collection: str
    query: 'Query'
    arguments: Dict[str, 'Argument']
    collection_relationships: Dict[str, 'Relationship']
    variables: Optional[List[Dict[str, Any]]] = None


class MutationRequest(RootModel):
    operations: List['MutationOperation']
    collection_relationships: Dict[str, 'Relationship']


class MutationResponse(RootModel):
    operation_results: List['MutationOperationResults']


class ExplainResponse(RootModel):
    details: Dict[str, str]


class ErrorResponse(RootModel):
    message: str
    details: Dict[str, Any]


class ValidateResponse(RootModel):
    schema: SchemaResponse
    capabilities: CapabilitiesResponse
    resolved_configuration: str


VERSION = "0.1.6"
